@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # ------------------------------------------------------------------------------
 #
-#   Copyright 2021 Valory AG
+#   Copyright 2021-2022 Valory AG
 #   Copyright 2018-2019 Fetch.AI Limited
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
@@ -24,6 +24,7 @@ import logging
 import math
 import threading
 import warnings
+from copy import deepcopy
 from pathlib import Path
 from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
 from uuid import uuid4
@@ -39,10 +40,13 @@ from eth_keys import keys
 from eth_typing import HexStr
 from eth_utils.currency import from_wei, to_wei  # pylint: disable=import-error
 from lru import LRU  # type: ignore  # pylint: disable=no-name-in-module
+from requests import HTTPError
 from web3 import HTTPProvider, Web3
 from web3.datastructures import AttributeDict
+from web3.exceptions import ContractLogicError, SolidityError, TransactionNotFound
 from web3.gas_strategies.rpc import rpc_gas_price_strategy
-from web3.types import TxData, TxParams, TxReceipt, Wei
+from web3.middleware import geth_poa_middleware
+from web3.types import LatestBlockParam, TxData, TxParams, TxReceipt, Wei
 
 from aea.common import Address, JSONLike
 from aea.crypto.base import Crypto, FaucetApi, Helper, LedgerApi
@@ -63,7 +67,12 @@ DEFAULT_CURRENCY_DENOM = "wei"
 ETH_GASSTATION_URL = "https://ethgasstation.info/api/ethgasAPI.json"
 _ABI = "abi"
 _BYTECODE = "bytecode"
-
+EIP1559 = "eip1559"
+EIP1559_POLYGON = "eip1559_polygon"
+GAS_STATION = "gas_station"
+AVAILABLE_STRATEGIES = (EIP1559, GAS_STATION, EIP1559_POLYGON)
+SPEED_FAST = "fast"  # safeLow, standard, fast
+POLYGON_GAS_ENDPOINT = "https://gasstation-mainnet.matic.network/v2"
 MAX_GAS_FAST = 1500
 
 # How many blocks to consider for priority fee estimation
@@ -82,12 +91,10 @@ DEFAULT_PRIORITY_FEE = 3
 FALLBACK_ESTIMATE = {
     "maxFeePerGas": to_wei(20, "gwei"),
     "maxPriorityFeePerGas": to_wei(DEFAULT_PRIORITY_FEE, "gwei"),
-    "baseFee": None,
 }
 
 PRIORITY_FEE_INCREASE_BOUNDARY = 200  # percentage
 
-
 DEFAULT_EIP1559_STRATEGY = {
     "max_gas_fast": MAX_GAS_FAST,
     "fee_history_blocks": FEE_HISTORY_BLOCKS,
@@ -98,9 +105,23 @@ DEFAULT_EIP1559_STRATEGY = {
     "priority_fee_increase_boundary": PRIORITY_FEE_INCREASE_BOUNDARY,
 }
 
+DEFAULT_EIP1559_STRATEGY_POLYGON = {
+    "gas_endpoint": POLYGON_GAS_ENDPOINT,
+    "speed": SPEED_FAST,
+    "fallback_estimate": FALLBACK_ESTIMATE,
+}
 
 DEFAULT_GAS_STATION_STRATEGY = {"gas_price_api_key": "", "gas_price_strategy": "fast"}
 
+DEFAULT_GAS_PRICE_STRATEGIES = {
+    EIP1559: DEFAULT_EIP1559_STRATEGY,
+    GAS_STATION: DEFAULT_GAS_STATION_STRATEGY,
+    EIP1559_POLYGON: DEFAULT_EIP1559_STRATEGY_POLYGON,
+}
+
+# The tip increase is the minimum required of 10%.
+TIP_INCREASE = 1.1
+
 
 def wei_to_gwei(number: Type[int]) -> Union[int, decimal.Decimal]:
     """Covert WEI to GWEI"""
@@ -136,17 +157,19 @@ def estimate_priority_fee(
     fee_history_blocks: int,
     fee_history_percentile: int,
     priority_fee_increase_boundary: int,
-) -> int:
+) -> Optional[int]:
     """Estimate priority fee from base fee."""
 
     if base_fee_gwei < priority_fee_estimation_trigger:
         return default_priority_fee
 
     fee_history = web3_object.eth.fee_history(
-        fee_history_blocks, block_number, [fee_history_percentile]
+        fee_history_blocks, block_number, [fee_history_percentile]  # type: ignore
     )
 
-    rewards = sorted([reward for reward in fee_history["reward"] if reward > 0])
+    # This is going to break if more percentiles are introduced in the future,
+    # i.e., `fee_history_percentile` param becomes a `List[int]`.
+    rewards = sorted([reward[0] for reward in fee_history["reward"] if reward[0] > 0])
     if len(rewards) == 0:
         return None
 
@@ -185,9 +208,11 @@ def get_gas_price_strategy_eip1559(
         transaction_params: TxParams,  # pylint: disable=unused-argument
     ) -> Dict[str, Wei]:
         """
-        Get gas price from Eth Gas Station api.
+        Get gas price using EIP1559.
+
+        Visit `https://github.com/ethereum/EIPs/blob/master/EIPS/eip-1559.md`
+        for more information.
 
-        Visit `https://docs.ethgasstation.info/gas-price` for documentation.
         :param web3: web3 instance
         :param transaction_params: transaction parameters
         :return: dictionary containing gas price strategy
@@ -200,7 +225,7 @@ def get_gas_price_strategy_eip1559(
 
         estimated_priority_fee = estimate_priority_fee(
             web3,
-            base_fee_gwei,
+            cast(int, base_fee_gwei),
             block_number,
             priority_fee_estimation_trigger=priority_fee_estimation_trigger,
             default_priority_fee=default_priority_fee,
@@ -218,7 +243,7 @@ def get_gas_price_strategy_eip1559(
         max_priority_fee_per_gas = max(
             estimated_priority_fee, to_wei(default_priority_fee, "gwei")
         )
-        multiplier = get_base_fee_multiplier(base_fee_gwei)
+        multiplier = get_base_fee_multiplier(cast(int, base_fee_gwei))
 
         potential_max_fee = base_fee * multiplier
         max_fee_per_gas = (
@@ -241,6 +266,32 @@ def get_gas_price_strategy_eip1559(
     return eip1559_price_strategy
 
 
+def get_gas_price_strategy_eip1559_polygon(
+    gas_endpoint: str,
+    fallback_estimate: Dict[str, Optional[int]],
+    speed: Optional[str] = SPEED_FAST,
+) -> Callable[[Any, Any], Dict[str, Wei]]:
+    """Get the gas price strategy."""
+
+    def eip1559_price_strategy(
+        web3: Web3,  # pylint: disable=redefined-outer-name,unused-argument
+        transaction_params: TxParams,  # pylint: disable=unused-argument
+    ) -> Dict[str, Wei]:
+        try:
+            response = requests.get(gas_endpoint)
+            if response.status_code == 200:
+                data = response.json()[speed]
+                return {
+                    "maxFeePerGas": Wei(to_wei(data["maxFee"], "gwei")),
+                    "maxPriorityFeePerGas": Wei(to_wei(data["maxPriorityFee"], "gwei")),
+                }
+            return fallback_estimate
+        except requests.exceptions.RequestException:
+            return fallback_estimate
+
+    return eip1559_price_strategy
+
+
 def rpc_gas_price_strategy_wrapper(
     web3: Web3, transaction_params: TxParams  # pylint: disable=redefined-outer-name
 ) -> Dict[str, Wei]:
@@ -609,7 +660,11 @@ class EthereumHelper(Helper):
 
     @staticmethod
     def is_transaction_valid(
-        tx: dict, seller: Address, client: Address, tx_nonce: str, amount: int,
+        tx: dict,
+        seller: Address,
+        client: Address,
+        tx_nonce: str,
+        amount: int,
     ) -> bool:
         """
         Check whether a transaction is valid or not.
@@ -742,8 +797,9 @@ class EthereumApi(LedgerApi, EthereumHelper):
     identifier = _ETHEREUM
 
     _gas_price_strategy_callables: Dict[str, Callable] = {
-        "gas_station": get_gas_price_strategy,
-        "eip1559": get_gas_price_strategy_eip1559,
+        GAS_STATION: get_gas_price_strategy,
+        EIP1559: get_gas_price_strategy_eip1559,
+        EIP1559_POLYGON: get_gas_price_strategy_eip1559_polygon,
     }
 
     def __init__(self, **kwargs: Any):
@@ -759,32 +815,51 @@ class EthereumApi(LedgerApi, EthereumHelper):
         self._is_gas_estimation_enabled = kwargs.pop("is_gas_estimation_enabled", False)
 
         self._default_gas_price_strategy: str = kwargs.pop(
-            "default_gas_price_strategy", "eip1559"
+            "default_gas_price_strategy", EIP1559
         )
+        if self._default_gas_price_strategy not in AVAILABLE_STRATEGIES:
+            raise ValueError(
+                f"Gas price strategy must be one of {AVAILABLE_STRATEGIES}, provided: {self._default_gas_price_strategy}"
+            )  # pragma: nocover
+
         self._gas_price_strategies: Dict[str, Dict] = kwargs.pop(
-            "gas_price_strategies", {}
+            "gas_price_strategies", DEFAULT_GAS_PRICE_STRATEGIES
         )
 
+        self._poa_chain = kwargs.pop("poa_chain", False)
+        if self._poa_chain:
+            # https://web3py.readthedocs.io/en/stable/middleware.html#geth-style-proof-of-authority
+            self._api.middleware_onion.inject(
+                geth_poa_middleware, name="geth_poa_middleware", layer=0
+            )
+            _default_logger.info(
+                "EthereumApi has been configured with Proof of Authority chain support"
+            )
+
     @property
     def api(self) -> Web3:
         """Get the underlying API object."""
         return self._api
 
-    def get_balance(self, address: Address) -> Optional[int]:
+    def get_balance(
+        self, address: Address, raise_on_try: bool = False
+    ) -> Optional[int]:
         """Get the balance of a given account."""
-        return self._try_get_balance(address)
+        return self._try_get_balance(address, raise_on_try=raise_on_try)
 
     @try_decorator("Unable to retrieve balance: {}", logger_method="warning")
-    def _try_get_balance(self, address: Address) -> Optional[int]:
+    def _try_get_balance(self, address: Address, **_kwargs: Any) -> Optional[int]:
         """Get the balance of a given account."""
         check_address = self._api.toChecksumAddress(address)
         return self._api.eth.get_balance(check_address)  # pylint: disable=no-member
 
     def get_state(
-        self, callable_name: str, *args: Any, **kwargs: Any
+        self, callable_name: str, *args: Any, raise_on_try: bool = False, **kwargs: Any
     ) -> Optional[JSONLike]:
         """Call a specified function on the ledger API."""
-        response = self._try_get_state(callable_name, *args, **kwargs)
+        response = self._try_get_state(
+            callable_name, *args, raise_on_try=raise_on_try, **kwargs
+        )
         return response
 
     @try_decorator("Unable to get state: {}", logger_method="warning")
@@ -793,6 +868,12 @@ class EthereumApi(LedgerApi, EthereumHelper):
     ) -> Optional[JSONLike]:
         """Try to call a function on the ledger API."""
 
+        if "raise_on_try" in kwargs:
+            logging.info(
+                f"popping `raise_on_try` from {self.__class__.__name__}.get_state kwargs"
+            )
+            kwargs.pop("raise_on_try")
+
         function = getattr(self._api.eth, callable_name)
         response = function(*args, **kwargs)
 
@@ -820,6 +901,7 @@ class EthereumApi(LedgerApi, EthereumHelper):
         gas_price: Optional[str] = None,
         gas_price_strategy: Optional[str] = None,
         gas_price_strategy_extra_config: Optional[Dict] = None,
+        raise_on_try: bool = False,
         **kwargs: Any,
     ) -> Optional[JSONLike]:
         """
@@ -836,6 +918,7 @@ class EthereumApi(LedgerApi, EthereumHelper):
         :param gas_price: the gas price (in Wei)
         :param gas_price_strategy: the gas price strategy to be used.
         :param gas_price_strategy_extra_config: extra config for gas price strategy.
+        :param raise_on_try: whether the method will raise or log on error
         :param kwargs: keyword arguments
         :return: the transfer transaction
         """
@@ -843,7 +926,10 @@ class EthereumApi(LedgerApi, EthereumHelper):
         chain_id = chain_id if chain_id is not None else self._chain_id
         destination_address = self._api.toChecksumAddress(destination_address)
         sender_address = self._api.toChecksumAddress(sender_address)
-        nonce = self._try_get_transaction_count(sender_address)
+        nonce = self._try_get_transaction_count(
+            sender_address,
+            raise_on_try=raise_on_try,
+        )
         if nonce is None:
             return transaction
         transaction = {
@@ -859,7 +945,7 @@ class EthereumApi(LedgerApi, EthereumHelper):
 
         if max_fee_per_gas is not None:
             max_priority_fee_per_gas = (
-                self._try_get_max_priority_fee()
+                self._try_get_max_priority_fee(raise_on_try=raise_on_try)
                 if max_priority_fee_per_gas is None
                 else max_priority_fee_per_gas
             )
@@ -874,8 +960,10 @@ class EthereumApi(LedgerApi, EthereumHelper):
             transaction.update({"gasPrice": gas_price})
 
         if gas_price is None and max_fee_per_gas is None:
-            gas_pricing = self._try_get_gas_pricing(
-                gas_price_strategy, gas_price_strategy_extra_config
+            gas_pricing = self.try_get_gas_pricing(
+                gas_price_strategy,
+                gas_price_strategy_extra_config,
+                raise_on_try=raise_on_try,
             )
             if gas_pricing is None:
                 return transaction  # pragma: nocover
@@ -883,44 +971,105 @@ class EthereumApi(LedgerApi, EthereumHelper):
 
         return transaction
 
-    @try_decorator("Unable to retrieve gas price: {}", logger_method="warning")
-    def _try_get_gas_pricing(
+    def _get_gas_price_strategy(
         self,
         gas_price_strategy: Optional[str] = None,
         extra_config: Optional[Dict] = None,
-    ) -> Optional[Dict[str, int]]:
-        """Try get the gas price based on the provided strategy."""
+    ) -> Optional[Tuple[str, Callable]]:
+        """
+        Returns parameters for gas price callable.
 
+        Note: The priority of gas price callable will be
+        `extra_config(Runtime params) > self._gas_price_strategies (Set using config file.) > DEFAULT_GAS_PRICE_STRATEGIES (Default values.)`
+
+        :param gas_price_strategy: name of the gas price strategy.
+        :param extra_config: gas price strategy getter parameters.
+        :return: gas price strategy's name and callable.
+        """
         gas_price_strategy = (
             gas_price_strategy
             if gas_price_strategy is not None
             else self._default_gas_price_strategy
         )
-        _default_logger.debug(f"Using strategy: {gas_price_strategy}")
-        extra_config = extra_config or {}
-        gas_price_strategy_getter = self._gas_price_strategy_callables.get(
-            gas_price_strategy, None
-        )
-        if gas_price_strategy_getter is None:
+        if gas_price_strategy not in AVAILABLE_STRATEGIES:  # pragma: nocover
             _default_logger.debug(
-                "No strategy found! Check configuration of ethereum-ledger-api."
+                f"Gas price strategy must be one of {AVAILABLE_STRATEGIES}, provided: {self._default_gas_price_strategy}"
             )
-            return None  # pragma: nocover
-        kwargs = self._gas_price_strategies.get(gas_price_strategy, {}).copy()
-        kwargs.update(extra_config)
-        gas_price_strategy_callable = gas_price_strategy_getter(**kwargs)
+            return None
+
+        _default_logger.debug(f"Using strategy: {gas_price_strategy}")
+        gas_price_strategy_getter = self._gas_price_strategy_callables[
+            gas_price_strategy
+        ]
+
+        parameters = cast(dict, DEFAULT_GAS_PRICE_STRATEGIES.get(gas_price_strategy))
+        parameters.update(self._gas_price_strategies.get(gas_price_strategy, {}))
+        parameters.update(extra_config or {})
+        return gas_price_strategy, gas_price_strategy_getter(**parameters)
+
+    @staticmethod
+    def __reprice(old_price: Wei) -> Wei:
+        return Wei(math.ceil(old_price * TIP_INCREASE))
+
+    @try_decorator("Unable to retrieve gas price: {}", logger_method="warning")
+    def try_get_gas_pricing(
+        self,
+        gas_price_strategy: Optional[str] = None,
+        extra_config: Optional[Dict] = None,
+        old_price: Optional[Dict[str, Wei]] = None,
+        **_kwargs: Any,
+    ) -> Optional[Dict[str, Wei]]:
+        """
+        Try get the gas price based on the provided strategy.
+
+        :param gas_price_strategy: the gas price strategy to use, e.g., the EIP-1559 strategy.
+            Can be either `eip1559` or `gas_station`.
+        :param extra_config: gas price strategy getter parameters.
+        :param old_price: the old gas price params in case that we are trying to resubmit a transaction.
+        :param _kwargs: the keyword arguments. Possible kwargs are:
+            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
+        :return: a dictionary with the gas data.
+        """
+
+        retrieved_strategy = self._get_gas_price_strategy(
+            gas_price_strategy,
+            extra_config,
+        )
+        if retrieved_strategy is None:  # pragma: nocover
+            return None
+        gas_price_strategy, gas_price_strategy_callable = retrieved_strategy
 
         prior_strategy = self._api.eth.gasPriceStrategy
         try:
             self._api.eth.set_gas_price_strategy(gas_price_strategy_callable)
             gas_price = self._api.eth.generate_gas_price()
         finally:
-            if prior_strategy is not None:
-                self._api.eth.set_gas_price_strategy(prior_strategy)  # pragma: nocover
+            self._api.eth.set_gas_price_strategy(prior_strategy)  # pragma: nocover
+
+        if gas_price is None or old_price is None:
+            return gas_price
+
+        gas_price = cast(Dict[str, Wei], gas_price)
+        if gas_price_strategy in (EIP1559, EIP1559_POLYGON):
+            updated_max_fee_per_gas = self.__reprice(old_price["maxFeePerGas"])
+            updated_max_priority_fee_per_gas = self.__reprice(
+                old_price["maxPriorityFeePerGas"]
+            )
+
+            if gas_price["maxFeePerGas"] < updated_max_fee_per_gas:
+                gas_price["maxFeePerGas"] = updated_max_fee_per_gas
+                gas_price["maxPriorityFeePerGas"] = updated_max_priority_fee_per_gas
+
+        elif gas_price_strategy == GAS_STATION:
+            updated_gas_price = self.__reprice(old_price["gasPrice"])
+            gas_price["gasPrice"] = max(gas_price["gasPrice"], updated_gas_price)
+
         return gas_price
 
     @try_decorator("Unable to retrieve transaction count: {}", logger_method="warning")
-    def _try_get_transaction_count(self, address: Address) -> Optional[int]:
+    def _try_get_transaction_count(
+        self, address: Address, **_kwargs: Any
+    ) -> Optional[int]:
         """Try get the transaction count."""
         nonce = self._api.eth.get_transaction_count(  # pylint: disable=no-member
             self._api.toChecksumAddress(address)
@@ -942,27 +1091,58 @@ class EthereumApi(LedgerApi, EthereumHelper):
     @try_decorator("Unable to retrieve gas estimate: {}", logger_method="warning")
     def _try_get_gas_estimate(self, transaction: JSONLike) -> Optional[int]:
         """Try get the gas estimate."""
-        gas_estimate = self._api.eth.estimate_gas(  # pylint: disable=no-member
-            transaction=cast(TxParams, AttributeDictTranslator.from_dict(transaction))
-        )
+        gas_estimate: Optional[int] = None
+        transaction = deepcopy(transaction)
+        del transaction["gas"]
+        try:
+            gas_estimate = self._api.eth.estimate_gas(  # pylint: disable=no-member
+                transaction=cast(
+                    TxParams, AttributeDictTranslator.from_dict(transaction)
+                )
+            )
+        except (ContractLogicError, ValueError) as e:
+            _default_logger.warning(
+                f"Unable to estimate gas with default state , "
+                f"{type(e).__name__}: {e.__str__()}"
+            )
+            # gas estimation might fail when repricing txs
+            # to avoid effects of pending txs when estimating gas
+            # we can set the block identifier to "latest" block
+            # this might fail if the node doesn't support the `block_identifier` param
+            gas_estimate = self._api.eth.estimate_gas(  # pylint: disable=no-member
+                transaction=cast(
+                    TxParams, AttributeDictTranslator.from_dict(transaction)
+                ),
+                block_identifier=LatestBlockParam,
+            )
+
         return gas_estimate
 
-    def send_signed_transaction(self, tx_signed: JSONLike) -> Optional[str]:
+    def send_signed_transaction(
+        self, tx_signed: JSONLike, raise_on_try: bool = False
+    ) -> Optional[str]:
         """
         Send a signed transaction and wait for confirmation.
 
         :param tx_signed: the signed transaction
+        :param raise_on_try: whether the method will raise or log on error
         :return: tx_digest, if present
         """
-        tx_digest = self._try_send_signed_transaction(tx_signed)
+        tx_digest = self._try_send_signed_transaction(
+            tx_signed, raise_on_try=raise_on_try
+        )
         return tx_digest
 
     @try_decorator("Unable to send transaction: {}", logger_method="warning")
-    def _try_send_signed_transaction(self, tx_signed: JSONLike) -> Optional[str]:
+    def _try_send_signed_transaction(
+        self, tx_signed: JSONLike, **_kwargs: Any
+    ) -> Optional[str]:
         """
         Try send a signed transaction.
 
         :param tx_signed: the signed transaction
+        :param _kwargs: the keyword arguments. Possible kwargs are:
+            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
         :return: tx_digest, if present
         """
         signed_transaction = SignedTransactionTranslator.from_dict(tx_signed)
@@ -975,24 +1155,42 @@ class EthereumApi(LedgerApi, EthereumHelper):
         )
         return tx_digest
 
-    def get_transaction_receipt(self, tx_digest: str) -> Optional[JSONLike]:
+    def get_transaction_receipt(
+        self, tx_digest: str, raise_on_try: bool = False
+    ) -> Optional[JSONLike]:
         """
         Get the transaction receipt for a transaction digest.
 
         :param tx_digest: the digest associated to the transaction.
+        :param raise_on_try: whether the method will raise or log on error
         :return: the tx receipt, if present
         """
-        tx_receipt = self._try_get_transaction_receipt(tx_digest)
+        tx_receipt = self._try_get_transaction_receipt(
+            tx_digest,
+            raise_on_try=raise_on_try,
+        )
+
+        if tx_receipt is not None and not bool(tx_receipt["status"]):
+            tx = self.get_transaction(tx_digest, raise_on_try=raise_on_try)
+            tx_receipt["revert_reason"] = self._try_get_revert_reason(
+                tx,
+                raise_on_try=raise_on_try,
+            )
+
         return tx_receipt
 
     @try_decorator(
         "Error when attempting getting tx receipt: {}", logger_method="debug"
     )
-    def _try_get_transaction_receipt(self, tx_digest: str) -> Optional[JSONLike]:
+    def _try_get_transaction_receipt(
+        self, tx_digest: str, **_kwargs: Any
+    ) -> Optional[JSONLike]:
         """
         Try get the transaction receipt.
 
         :param tx_digest: the digest associated to the transaction.
+        :param _kwargs: the keyword arguments. Possible kwargs are:
+            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
         :return: the tx receipt, if present
         """
         tx_receipt = self._api.eth.get_transaction_receipt(  # pylint: disable=no-member
@@ -1000,22 +1198,31 @@ class EthereumApi(LedgerApi, EthereumHelper):
         )
         return AttributeDictTranslator.to_dict(tx_receipt)
 
-    def get_transaction(self, tx_digest: str) -> Optional[JSONLike]:
+    def get_transaction(
+        self,
+        tx_digest: str,
+        raise_on_try: bool = False,
+    ) -> Optional[JSONLike]:
         """
         Get the transaction for a transaction digest.
 
         :param tx_digest: the digest associated to the transaction.
+        :param raise_on_try: whether the method will raise or log on error
         :return: the tx, if present
         """
-        tx = self._try_get_transaction(tx_digest)
+        tx = self._try_get_transaction(tx_digest, raise_on_try=raise_on_try)
         return tx
 
     @try_decorator("Error when attempting getting tx: {}", logger_method="debug")
-    def _try_get_transaction(self, tx_digest: str) -> Optional[JSONLike]:
+    def _try_get_transaction(
+        self, tx_digest: str, **_kwargs: Any
+    ) -> Optional[JSONLike]:
         """
         Get the transaction.
 
         :param tx_digest: the transaction digest.
+        :param _kwargs: the keyword arguments. Possible kwargs are:
+            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
         :return: the tx, if found
         """
         tx = self._api.eth.get_transaction(
@@ -1023,6 +1230,38 @@ class EthereumApi(LedgerApi, EthereumHelper):
         )  # pylint: disable=no-member
         return AttributeDictTranslator.to_dict(tx)
 
+    @try_decorator(
+        "Error when attempting getting tx revert reason: {}", logger_method="debug"
+    )
+    def _try_get_revert_reason(self, tx: TxData, **_kwargs: Any) -> str:
+        """Try to check the revert reason of a transaction.
+
+        :param tx: the transaction for which we want to get the revert reason.
+        :param _kwargs: the keyword arguments. Possible kwargs are:
+            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
+        :return: the revert reason message.
+        """
+        # build a new transaction to replay:
+        replay_tx = {
+            "to": tx["to"],
+            "from": tx["from"],
+            "value": tx["value"],
+            "data": tx["input"],
+        }
+
+        try:
+            # replay the transaction on the provider
+            self.api.eth.call(replay_tx, tx["blockNumber"] - 1)
+        except SolidityError as e:
+            # execution reverted exception
+            return str(e)
+        except HTTPError as e:
+            # http exception
+            raise e
+        else:
+            # given tx not reverted
+            raise ValueError(f"The given transaction has not been reverted!\ntx: {tx}")
+
     def get_contract_instance(
         self, contract_interface: Dict[str, str], contract_address: Optional[str] = None
     ) -> Any:
@@ -1035,7 +1274,8 @@ class EthereumApi(LedgerApi, EthereumHelper):
         """
         if contract_address is None:
             instance = self.api.eth.contract(
-                abi=contract_interface[_ABI], bytecode=contract_interface[_BYTECODE],
+                abi=contract_interface[_ABI],
+                bytecode=contract_interface[_BYTECODE],
             )
         else:
             _contract_address = self.api.toChecksumAddress(contract_address)
@@ -1050,13 +1290,7 @@ class EthereumApi(LedgerApi, EthereumHelper):
         self,
         contract_interface: Dict[str, str],
         deployer_address: Address,
-        value: int = 0,
-        gas: Optional[int] = None,
-        max_fee_per_gas: Optional[int] = None,
-        max_priority_fee_per_gas: Optional[str] = None,
-        gas_price: Optional[str] = None,
-        gas_price_strategy: Optional[str] = None,
-        gas_price_strategy_extra_config: Optional[Dict] = None,
+        raise_on_try: bool = False,
         **kwargs: Any,
     ) -> Optional[JSONLike]:
         """
@@ -1064,19 +1298,43 @@ class EthereumApi(LedgerApi, EthereumHelper):
 
         :param contract_interface: the contract interface.
         :param deployer_address: The address that will deploy the contract.
-        :param value: value to send to contract (in Wei)
-        :param gas: the gas to be used (in Wei)
-        :param max_fee_per_gas: maximum amount you’re willing to pay, inclusive of `baseFeePerGas` and `maxPriorityFeePerGas`. The difference between `maxFeePerGas` and `baseFeePerGas + maxPriorityFeePerGas` is refunded  (in Wei).
-        :param max_priority_fee_per_gas: the part of the fee that goes to the miner (in Wei).
-        :param gas_price: the gas price (in Wei)
-        :param gas_price_strategy: the gas price strategy to be used.
-        :param gas_price_strategy_extra_config: extra config for gas price strategy..
+        :param raise_on_try: whether the method will raise or log on error
         :param kwargs: keyword arguments
         :return: the transaction dictionary.
         """
+
+        # value to send to contract (in Wei)
+        value: int = kwargs.pop("value", 0)
+
+        # the gas to be used (in Wei)
+        gas: Optional[int] = kwargs.pop("gas", None)
+
+        # maximum amount you’re willing to pay, inclusive of `baseFeePerGas` and
+        # `maxPriorityFeePerGas`. The difference between `maxFeePerGas` and
+        # `baseFeePerGas + maxPriorityFeePerGas` is refunded  (in Wei).
+        max_fee_per_gas: Optional[int] = kwargs.pop("max_fee_per_gas", None)
+
+        # the part of the fee that goes to the miner (in Wei).
+        max_priority_fee_per_gas: Optional[str] = kwargs.pop(
+            "max_priority_fee_per_gas", None
+        )
+
+        # the gas price (in Wei)
+        gas_price: Optional[str] = kwargs.pop("gas_price", None)
+
+        # the gas price strategy to be used.
+        gas_price_strategy: Optional[str] = kwargs.pop("gas_price_strategy", None)
+
+        # extra config for gas price strategy.
+        gas_price_strategy_extra_config: Optional[Dict] = kwargs.pop(
+            "gas_price_strategy_extra_config", None
+        )
+
         transaction: Optional[JSONLike] = None
         _deployer_address = self.api.toChecksumAddress(deployer_address)
-        nonce = self._try_get_transaction_count(_deployer_address)
+        nonce = self._try_get_transaction_count(
+            _deployer_address, raise_on_try=raise_on_try
+        )
         if nonce is None:
             return transaction
         instance = self.get_contract_instance(contract_interface)
@@ -1086,7 +1344,7 @@ class EthereumApi(LedgerApi, EthereumHelper):
         }
         if max_fee_per_gas is not None:
             max_priority_fee_per_gas = (
-                self._try_get_max_priority_fee()
+                self._try_get_max_priority_fee(raise_on_try=raise_on_try)
                 if max_priority_fee_per_gas is None
                 else max_priority_fee_per_gas
             )
@@ -1103,15 +1361,19 @@ class EthereumApi(LedgerApi, EthereumHelper):
             transaction.update({"gasPrice": gas_price})
 
         if gas_price is None and max_fee_per_gas is None:
-            gas_pricing = self._try_get_gas_pricing(
-                gas_price_strategy, gas_price_strategy_extra_config
+            gas_pricing = self.try_get_gas_pricing(
+                gas_price_strategy,
+                gas_price_strategy_extra_config,
+                raise_on_try=raise_on_try,
             )
 
             if gas_pricing is None:
                 return None  # pragma: nocover
+
             transaction.update(gas_pricing)
 
         transaction = instance.constructor(**kwargs).buildTransaction(transaction)
+
         if transaction is None:
             return None  # pragma: nocover
         transaction.pop("to", None)  # only 'from' address, don't insert 'to' address!
@@ -1123,7 +1385,7 @@ class EthereumApi(LedgerApi, EthereumHelper):
         return transaction
 
     @try_decorator("Unable to retrieve max_priority_fee: {}", logger_method="warning")
-    def _try_get_max_priority_fee(self) -> str:
+    def _try_get_max_priority_fee(self, **_kwargs: Any) -> str:
         """Try get the gas estimate."""
         return cast(str, self.api.eth.max_priority_fee)
 
@@ -1137,6 +1399,113 @@ class EthereumApi(LedgerApi, EthereumHelper):
         """
         return Web3.isAddress(address)
 
+    @classmethod
+    def contract_method_call(
+        cls,
+        contract_instance: Any,
+        method_name: str,
+        **method_args: Any,
+    ) -> Optional[JSONLike]:
+        """Call a contract's method
+
+        :param contract_instance: the contract to use
+        :param method_name: the contract method to call
+        :param method_args: the contract call parameters
+        :return: the call result
+        """
+        method = getattr(contract_instance.functions, method_name)
+        result = method(**method_args).call()
+        return result
+
+    def build_transaction(  # pylint: disable=too-many-arguments
+        self,
+        contract_instance: Any,
+        method_name: str,
+        method_args: Optional[Dict[Any, Any]],
+        tx_args: Optional[Dict[Any, Any]],
+        raise_on_try: bool = False,
+    ) -> Optional[JSONLike]:
+        """Prepare a transaction
+
+        :param contract_instance: the contract to use
+        :param method_name: the contract method to call
+        :param method_args: the contract parameters
+        :param tx_args: the transaction parameters
+        :param raise_on_try: whether the method will raise or log on error
+        :return: the transaction
+        """
+
+        if method_args is None:
+            raise ValueError("Argument 'method_args' cannot be 'None'.")
+
+        method = getattr(contract_instance.functions, method_name)
+        tx = method(**cast(Dict, method_args))
+
+        if tx_args is None:
+            raise ValueError("Argument 'tx_args' cannot be 'None'.")
+
+        tx_args = cast(Dict, tx_args)
+
+        nonce = self.api.eth.get_transaction_count(tx_args["sender_address"])
+        tx_params = {
+            "nonce": nonce,
+            "value": tx_args["value"] if "value" in tx_args else 0,
+            "gas": 1,  # set this as a placeholder to avoid estimation on buildTransaction()
+        }
+
+        # Parameter camel-casing due to contract api requirements
+        for field in [
+            "gas",
+            "gasPrice",
+            "maxFeePerGas",
+            "maxPriorityFeePerGas",
+        ]:
+            if field in tx_args and tx_args[field] is not None:
+                tx_params[field] = tx_args[field]
+
+        if (
+            "gasPrice" not in tx_params
+            and "maxFeePerGas" not in tx_params
+            and "maxPriorityFeePerGas" not in tx_params
+        ):
+            gas_data = self.try_get_gas_pricing(
+                old_price=tx_args.get("old_price"), raise_on_try=raise_on_try
+            )
+            if gas_data:
+                tx_params.update(gas_data)  # pragma: nocover
+
+        tx = tx.buildTransaction(tx_params)
+        if self._is_gas_estimation_enabled:
+            tx = self.update_with_gas_estimate(tx)
+
+        return tx
+
+    def get_transaction_transfer_logs(  # pylint: disable=too-many-arguments,too-many-locals
+        self,
+        contract_instance: Any,
+        tx_hash: str,
+        target_address: Optional[str] = None,
+    ) -> Optional[JSONLike]:
+        """
+        Get all transfer events derived from a transaction.
+
+        :param contract_instance: the contract
+        :param tx_hash: the transaction hash
+        :param target_address: optional address to filter tranfer events to just those that affect it
+        :return: the transfer logs
+        """
+        try:
+            tx_receipt = self.api.eth.get_transaction_receipt(tx_hash)
+            if tx_receipt is None:
+                raise ValueError  # pragma: nocover
+
+        except (TransactionNotFound, ValueError):  # pragma: nocover
+            return dict(logs=[])
+
+        transfer_logs = contract_instance.events.Transfer().processReceipt(tx_receipt)
+
+        return dict(logs=transfer_logs)
+
 
 class EthereumFaucetApi(FaucetApi):
     """Ethereum testnet faucet API."""
