@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # ------------------------------------------------------------------------------
 #
-#   Copyright 2021 Valory AG
+#   Copyright 2021-2022 Valory AG
 #   Copyright 2018-2019 Fetch.AI Limited
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
@@ -23,10 +23,11 @@ import hashlib
 import logging
 import math
 import random
+import re
 import tempfile
 import time
 from pathlib import Path
-from typing import Dict, cast
+from typing import Dict, Generator, Optional, Tuple, Union, cast
 from unittest import mock
 from unittest.mock import MagicMock, patch
 
@@ -46,10 +47,17 @@ from aea_ledger_ethereum import (
 from aea_ledger_ethereum.ethereum import (
     DEFAULT_EIP1559_STRATEGY,
     DEFAULT_GAS_STATION_STRATEGY,
+    EIP1559,
+    EIP1559_POLYGON,
+    GAS_STATION,
+    TIP_INCREASE,
 )
 from web3 import Web3
 from web3._utils.request import _session_cache as session_cache
+from web3.datastructures import AttributeDict
+from web3.exceptions import SolidityError
 
+from aea.common import JSONLike
 from aea.crypto.helpers import DecryptError, KeyIsIncorrect
 
 from tests.conftest import DEFAULT_GANACHE_CHAIN_ID, MAX_FLAKY_RERUNS, ROOT_DIR
@@ -72,7 +80,8 @@ def get_history_data(n_blocks: int, base_multiplier: int = 100) -> Dict:
     return {
         "oldestBlock": 1,
         "reward": [
-            math.ceil(random.random() * base_multiplier) * 1e1 for _ in range(n_blocks)
+            [math.ceil(random.random() * base_multiplier) * 1e1]
+            for _ in range(n_blocks)
         ],
         "baseFeePerGas": [
             math.ceil(random.random() * base_multiplier) * 1e9 for _ in range(n_blocks)
@@ -185,6 +194,12 @@ def test_api_creation(ethereum_testnet_config):
     assert EthereumApi(**ethereum_testnet_config), "Failed to initialise the api"
 
 
+def test_api_creation_poa(ethereum_testnet_config):
+    """Test api instantiation with the poa flag enabled."""
+    ethereum_testnet_config["poa_chain"] = True
+    assert EthereumApi(**ethereum_testnet_config), "Failed to initialise the api"
+
+
 def test_api_none(ethereum_testnet_config):
     """Test the "api" of the cryptoApi is none."""
     eth_api = EthereumApi(**ethereum_testnet_config)
@@ -225,31 +240,33 @@ def test_get_state(ethereum_testnet_config, ganache):
     assert "number" in block, "response to get_block() does not contain 'number'"
 
 
-@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
-@pytest.mark.integration
-@pytest.mark.ledger
-def test_construct_sign_and_submit_transfer_transaction(
-    ethereum_testnet_config, ganache, ethereum_private_key_file
-):
-    """Test the construction, signing and submitting of a transfer transaction."""
-    account = EthereumCrypto(private_key_path=ethereum_private_key_file)
-    ec2 = EthereumCrypto()
-    ethereum_api = EthereumApi(**ethereum_testnet_config)
+def _wait_get_receipt(
+    ethereum_api: EthereumApi, transaction_digest: str
+) -> Tuple[Optional[JSONLike], bool]:
+    transaction_receipt = None
+    not_settled = True
+    elapsed_time = 0
+    time_to_wait = 40
+    sleep_time = 2
+    while not_settled and elapsed_time < time_to_wait:
+        elapsed_time += sleep_time
+        time.sleep(sleep_time)
+        transaction_receipt = ethereum_api.get_transaction_receipt(transaction_digest)
+        if transaction_receipt is None:
+            continue
+        is_settled = ethereum_api.is_transaction_settled(transaction_receipt)
+        not_settled = not is_settled
 
-    amount = 40000
-    tx_nonce = ethereum_api.generate_tx_nonce(ec2.address, account.address)
-    max_priority_fee_per_gas = 1_000_000_000
-    max_fee_per_gas = 1_000_000_000
-    transfer_transaction = ethereum_api.get_transfer_transaction(
-        sender_address=account.address,
-        destination_address=ec2.address,
-        amount=amount,
-        tx_fee=30000,
-        tx_nonce=tx_nonce,
-        chain_id=DEFAULT_GANACHE_CHAIN_ID,
-        max_priority_fee_per_gas=max_priority_fee_per_gas,
-        max_fee_per_gas=max_fee_per_gas,
-    )
+    return transaction_receipt, not not_settled
+
+
+def _construct_and_settle_tx(
+    ethereum_api: EthereumApi,
+    account: EthereumCrypto,
+    tx_params: dict,
+) -> Tuple[str, JSONLike, bool]:
+    """Construct and settle a transaction."""
+    transfer_transaction = ethereum_api.get_transfer_transaction(**tx_params)
     assert (
         isinstance(transfer_transaction, dict) and len(transfer_transaction) == 8
     ), "Incorrect transfer_transaction constructed."
@@ -262,22 +279,47 @@ def test_construct_sign_and_submit_transfer_transaction(
     transaction_digest = ethereum_api.send_signed_transaction(signed_transaction)
     assert transaction_digest is not None, "Failed to submit transfer transaction!"
 
-    not_settled = True
-    elapsed_time = 0
-    while not_settled and elapsed_time < 20:
-        elapsed_time += 1
-        time.sleep(2)
-        transaction_receipt = ethereum_api.get_transaction_receipt(transaction_digest)
-        if transaction_receipt is None:
-            continue
-        is_settled = ethereum_api.is_transaction_settled(transaction_receipt)
-        not_settled = not is_settled
+    transaction_receipt, is_settled = _wait_get_receipt(
+        ethereum_api, transaction_digest
+    )
+
     assert transaction_receipt is not None, "Failed to retrieve transaction receipt."
+
+    return transaction_digest, transaction_receipt, is_settled
+
+
+@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
+@pytest.mark.integration
+@pytest.mark.ledger
+def test_construct_sign_and_submit_transfer_transaction(
+    ethereum_testnet_config, ganache, ethereum_private_key_file
+):
+    """Test the construction, signing and submitting of a transfer transaction."""
+    account = EthereumCrypto(private_key_path=ethereum_private_key_file)
+    ec2 = EthereumCrypto()
+    ethereum_api = EthereumApi(**ethereum_testnet_config)
+
+    tx_params = {
+        "sender_address": account.address,
+        "destination_address": ec2.address,
+        "amount": 40000,
+        "tx_fee": 30000,
+        "tx_nonce": ethereum_api.generate_tx_nonce(ec2.address, account.address),
+        "chain_id": DEFAULT_GANACHE_CHAIN_ID,
+        "max_priority_fee_per_gas": 1_000_000_000,
+        "max_fee_per_gas": 1_000_000_000,
+    }
+
+    transaction_digest, transaction_receipt, is_settled = _construct_and_settle_tx(
+        ethereum_api,
+        account,
+        tx_params,
+    )
     assert is_settled, "Failed to verify tx!"
 
     tx = ethereum_api.get_transaction(transaction_digest)
     is_valid = ethereum_api.is_transaction_valid(
-        tx, ec2.address, account.address, tx_nonce, amount
+        tx, ec2.address, account.address, tx_params["tx_nonce"], tx_params["amount"]
     )
     assert is_valid, "Failed to settle tx correctly!"
     assert tx != transaction_receipt, "Should not be same!"
@@ -434,7 +476,11 @@ def test_gas_price_strategy_eip1559() -> None:
     )
 
     fee_history_mock = mock.patch.object(
-        web3.eth, "fee_history", return_value=get_history_data(n_blocks=5,),
+        web3.eth,
+        "fee_history",
+        return_value=get_history_data(
+            n_blocks=5,
+        ),
     )
 
     with get_block_mock:
@@ -442,7 +488,6 @@ def test_gas_price_strategy_eip1559() -> None:
             gas_stregy = callable_(web3, "tx_params")
 
     assert all([key in gas_stregy for key in ["maxFeePerGas", "maxPriorityFeePerGas"]])
-
     assert all([value > 1e8 for value in gas_stregy.values()])
 
 
@@ -457,7 +502,11 @@ def test_gas_price_strategy_eip1559_estimate_none() -> None:
     )
 
     fee_history_mock = mock.patch.object(
-        web3.eth, "fee_history", return_value=get_history_data(n_blocks=5,),
+        web3.eth,
+        "fee_history",
+        return_value=get_history_data(
+            n_blocks=5,
+        ),
     )
     with get_block_mock:
         with fee_history_mock:
@@ -469,8 +518,6 @@ def test_gas_price_strategy_eip1559_estimate_none() -> None:
 
     assert all([key in gas_stregy for key in ["maxFeePerGas", "maxPriorityFeePerGas"]])
 
-    assert gas_stregy["baseFee"] is None
-
 
 def test_gas_price_strategy_eip1559_fallback() -> None:
     """Test eip1559 based gas price strategy."""
@@ -485,7 +532,11 @@ def test_gas_price_strategy_eip1559_fallback() -> None:
     )
 
     fee_history_mock = mock.patch.object(
-        web3.eth, "fee_history", return_value=get_history_data(n_blocks=5,),
+        web3.eth,
+        "fee_history",
+        return_value=get_history_data(
+            n_blocks=5,
+        ),
     )
     with get_block_mock:
         with fee_history_mock:
@@ -497,8 +548,6 @@ def test_gas_price_strategy_eip1559_fallback() -> None:
 
     assert all([key in gas_stregy for key in ["maxFeePerGas", "maxPriorityFeePerGas"]])
 
-    assert gas_stregy["baseFee"] is None
-
 
 def test_gas_price_strategy_eth_gasstation():
     """Test the gas price strategy when using eth gasstation."""
@@ -588,3 +637,286 @@ def test_decrypt_error():
 def test_helper_get_contract_address():
     """Test EthereumHelper.get_contract_address."""
     assert EthereumHelper.get_contract_address({"contractAddress": "123"}) == "123"
+
+
+def test_contract_method_call():
+    """Test EthereumApi.contract_method_call."""
+
+    method_mock = MagicMock()
+    method_mock().call = MagicMock(return_value={"value": 0})
+
+    contract_instance = MagicMock()
+    contract_instance.functions.dummy_method = method_mock
+
+    result = EthereumApi.contract_method_call(
+        contract_instance=contract_instance, method_name="dummy_method", dummy_arg=1
+    )
+    assert result["value"] == 0
+
+
+def test_build_transaction(ethereum_testnet_config):
+    """Test EthereumApi.build_transaction."""
+
+    def pass_tx_params(tx_params):
+        return tx_params
+
+    tx_mock = MagicMock()
+    tx_mock.buildTransaction = pass_tx_params
+
+    method_mock = MagicMock(return_value=tx_mock)
+
+    contract_instance = MagicMock()
+    contract_instance.functions.dummy_method = method_mock
+
+    eth_api = EthereumApi(**ethereum_testnet_config)
+
+    with pytest.raises(
+        ValueError, match=re.escape("Argument 'method_args' cannot be 'None'.")
+    ):
+        eth_api.build_transaction(
+            contract_instance=contract_instance,
+            method_name="dummy_method",
+            method_args=None,
+            tx_args={},
+        )
+    with pytest.raises(
+        ValueError, match=re.escape("Argument 'tx_args' cannot be 'None'.")
+    ):
+        eth_api.build_transaction(
+            contract_instance=contract_instance,
+            method_name="dummy_method",
+            method_args={},
+            tx_args=None,
+        )
+
+    with mock.patch(
+        "web3.eth.Eth.get_transaction_count",
+        return_value=0,
+    ):
+        result = eth_api.build_transaction(
+            contract_instance=contract_instance,
+            method_name="dummy_method",
+            method_args={},
+            tx_args=dict(
+                sender_address="sender_address",
+                eth_value=0,
+                gas=0,
+                gasPrice=0,  # camel-casing due to contract api requirements
+                maxFeePerGas=0,  # camel-casing due to contract api requirements
+                maxPriorityFeePerGas=0,  # camel-casing due to contract api requirements
+            ),
+        )
+
+        assert result == dict(
+            nonce=0,
+            value=0,
+            gas=0,
+            gasPrice=0,
+            maxFeePerGas=0,
+            maxPriorityFeePerGas=0,
+        )
+
+        with mock.patch.object(
+            EthereumApi,
+            "try_get_gas_pricing",
+            return_value={"gas": 0},
+        ):
+            result = eth_api.build_transaction(
+                contract_instance=contract_instance,
+                method_name="dummy_method",
+                method_args={},
+                tx_args=dict(
+                    sender_address="sender_address",
+                    eth_value=0,
+                ),
+            )
+
+            assert result == dict(nonce=0, value=0, gas=0)
+
+
+def test_get_transaction_transfer_logs(ethereum_testnet_config):
+    """Test EthereumApi.get_transaction_transfer_logs."""
+    eth_api = EthereumApi(**ethereum_testnet_config)
+
+    dummy_receipt = {"logs": [{"topics": ["0x0", "0x0"]}]}
+
+    with mock.patch(
+        "web3.eth.Eth.get_transaction_receipt",
+        return_value=dummy_receipt,
+    ):
+        contract_instance = MagicMock()
+        contract_instance.events.Transfer().processReceipt.return_value = {"log": "log"}
+
+        result = eth_api.get_transaction_transfer_logs(
+            contract_instance=contract_instance,
+            tx_hash="dummy_hash",
+        )
+
+        assert result == dict(logs={"log": "log"})
+
+
+def test_get_transaction_transfer_logs_raise(ethereum_testnet_config):
+    """Test EthereumApi.get_transaction_transfer_logs."""
+    eth_api = EthereumApi(**ethereum_testnet_config)
+
+    with mock.patch(
+        "web3.eth.Eth.get_transaction_receipt",
+        return_value=None,
+    ):
+        contract_instance = MagicMock()
+        contract_instance.events.Transfer().processReceipt.return_value = {"log": "log"}
+
+        result = eth_api.get_transaction_transfer_logs(
+            contract_instance=contract_instance,
+            tx_hash="dummy_hash",
+        )
+
+        assert result == dict(logs=[])
+
+
+@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
+@pytest.mark.integration
+@pytest.mark.ledger
+def test_revert_reason(
+    ethereum_private_key_file: str,
+    ethereum_testnet_config: dict,
+    ganache: Generator,
+) -> None:
+    """Test the retrieval of the revert reason for a transaction."""
+    account = EthereumCrypto(private_key_path=ethereum_private_key_file)
+    ec2 = EthereumCrypto()
+    ethereum_api = EthereumApi(**ethereum_testnet_config)
+
+    tx_params = {
+        "sender_address": account.address,
+        "destination_address": ec2.address,
+        "amount": 40000,
+        "tx_fee": 30000,
+        "tx_nonce": 0,
+        "chain_id": DEFAULT_GANACHE_CHAIN_ID,
+        "max_priority_fee_per_gas": 1_000_000_000,
+        "max_fee_per_gas": 1_000_000_000,
+    }
+
+    with mock.patch(
+        "web3.eth.Eth.get_transaction_receipt",
+        return_value=AttributeDict({"status": 0}),
+    ):
+        with mock.patch(
+            "web3.eth.Eth.call",
+            side_effect=SolidityError("test revert reason"),
+        ):
+            _, transaction_receipt, is_settled = _construct_and_settle_tx(
+                ethereum_api,
+                account,
+                tx_params,
+            )
+
+            assert transaction_receipt["revert_reason"] == "test revert reason"
+
+
+@pytest.mark.parametrize(
+    "strategy",
+    (
+        {"name": EIP1559, "params": ("maxPriorityFeePerGas", "maxFeePerGas")},
+        {"name": GAS_STATION, "params": ("gasPrice",)},
+        {"name": EIP1559_POLYGON, "params": ("maxPriorityFeePerGas", "maxFeePerGas")},
+    ),
+)
+def test_try_get_gas_pricing(
+    strategy: Dict[str, Union[str, Tuple[str, ...]]],
+    ethereum_testnet_config: dict,
+    ganache: Generator,
+) -> None:
+    """Test `try_get_gas_pricing`."""
+    ethereum_api = EthereumApi(**ethereum_testnet_config)
+
+    # test gas pricing
+    gas_price = ethereum_api.try_get_gas_pricing(gas_price_strategy=strategy["name"])
+    assert set(strategy["params"]) == set(gas_price.keys())
+    assert all(
+        gas_price[param] > 0 and isinstance(gas_price[param], int)
+        for param in strategy["params"]
+    )
+
+    # test gas repricing
+    gas_reprice = ethereum_api.try_get_gas_pricing(
+        gas_price_strategy=strategy["name"], old_price=gas_price
+    )
+    assert all(
+        gas_reprice[param] > 0 and isinstance(gas_reprice[param], int)
+        for param in strategy["params"]
+    )
+    assert gas_reprice == {
+        gas_price_param: math.ceil(gas_price[gas_price_param] * TIP_INCREASE)
+        for gas_price_param in strategy["params"]
+    }, "The repricing was performed incorrectly!"
+
+
+@pytest.mark.parametrize(
+    "strategy",
+    ({"name": EIP1559_POLYGON, "params": ("maxPriorityFeePerGas", "maxFeePerGas")},),
+)
+def test_try_get_gas_pricing_poa(
+    strategy: Dict[str, Union[str, Tuple[str, ...]]],
+    polygon_testnet_config: dict,
+    ganache: Generator,
+) -> None:
+    """Test `try_get_gas_pricing` for a poa chain like Rinkeby."""
+    ethereum_api = EthereumApi(**polygon_testnet_config)
+    assert "geth_poa_middleware" in ethereum_api.api.middleware_onion.keys()
+
+    # test gas pricing
+    gas_price = ethereum_api.try_get_gas_pricing(gas_price_strategy=strategy["name"])
+    assert set(strategy["params"]) == set(gas_price.keys())
+    assert all(
+        gas_price[param] > 0 and isinstance(gas_price[param], int)
+        for param in strategy["params"]
+    )
+
+    # test gas repricing
+    gas_reprice = ethereum_api.try_get_gas_pricing(
+        gas_price_strategy=strategy["name"], old_price=gas_price
+    )
+    assert all(
+        gas_reprice[param] > 0 and isinstance(gas_reprice[param], int)
+        for param in strategy["params"]
+    )
+    assert gas_reprice == {
+        gas_price_param: math.ceil(gas_price[gas_price_param] * TIP_INCREASE)
+        for gas_price_param in strategy["params"]
+    }, "The repricing was performed incorrectly!"
+
+
+@pytest.mark.parametrize("mock_exception", (True, False))
+def test_gas_estimation(
+    mock_exception,
+    ethereum_testnet_config: dict,
+    ganache: Generator,
+    caplog,
+) -> None:
+    """Test gas estimation."""
+    ethereum_api = EthereumApi(**ethereum_testnet_config)
+    tx = {
+        "value": 0,
+        "chainId": 1337,
+        "from": "0xBcd4042DE499D14e55001CcbB24a551F3b954096",
+        "gas": 291661,
+        "maxPriorityFeePerGas": 3000000000,
+        "maxFeePerGas": 4000000000,
+        "to": "0x68FCdF52066CcE5612827E872c45767E5a1f6551",
+        "data": "",
+    }
+    with caplog.at_level(logging.DEBUG, logger="aea.crypto.ethereum._default_logger"):
+        with patch.object(ethereum_api._api.eth, "estimate_gas") as estimate_gas_mock:
+            if mock_exception:
+                # raise exception on first call only
+                estimate_gas_mock.side_effect = [
+                    ValueError("triggered exception"),
+                    None,
+                ]
+            ethereum_api.update_with_gas_estimate(tx)
+        if mock_exception:
+            assert (
+                "ValueError: triggered exception" in caplog.text
+            ), f"Cannot find message in output: {caplog.text}"
